/*
9vorbis -- an acme/win-friendly ogg-vorbis-player for Plan 9 from Bell-labs with playlist,shuffle, etc
Copyright (C) 2006 Sascha Retzki
released under the terms of the MITL
All rights reserved.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifndef _PLAN9_SOURCE
#define _PLAN9_SOURCE
#endif /* *sigh* */
#include <lib9.h>

#ifndef _BSD_EXTENSION
#define _BSD_EXTENSION
#endif /* *sigh* */
#include <bsd.h>

#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE
#endif /* *sigh* */
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifndef _REGEXP_EXTENSION
#define _REGEXP_EXTENSION
#endif /* *sigh* */
#include <regexp.h>

#include <time.h>

#include <vorbis/vorbisfile.h>



/*
The menue-commands are commented, just search for 
ec_menuecommand 
to jump there.
*/


enum
{
	/* return values (do never assume them to be 0|1...) */
	Error,
	Success,

	/* playmodes */
	Pplay,
	Ppause,
	Pstop
};


typedef struct Pls Pls;
struct Pls {
	char *filename;

	char *artist;
	char *title;

	Pls *next;
	Pls *prev;
};

Pls *plsroot, *plscurrent, *plslast;


long		audiorate;			/* The current sample-rate of audio(3) */


int		playmode = Pstop,	/* playmode indicates if we are currently playing, are stopped, etc */
		shuffle = 1,		/* Do we shuffle? (YES) */
		repeat = 0,		/* Do we repeat the playlist? */
		errorcount,		/* Counts all errors */
		skipper = 0,		/* skip N frames, used in playvorbis() to skip 'seconds' */
		announcements =1,	/* Do we want play() to announce the song? (YES) */
		pid,				/* our child pid */
		whattimeisit;		/* Used to announce the bitstream-position */


/* Prototypes... */
int execcmd(char *);


/************************************	misc		*********************************************/

void
announcer(void)
{
	if( plscurrent->artist[0] != 0 && plscurrent->title[0] != 0)
		printf("np: %s - %s (%s)\n",plscurrent->artist,plscurrent->title,plscurrent->filename);
	else
		printf("np:  %s\n",plscurrent->filename);
}

void
prepacme(void)
{
	int fd;

	if((fd = open("/dev/acme/tag",O_WRONLY | O_APPEND)) >= 0) {
		if(write(fd," play pause stop next prev pls",31) != 31)
			printf("errwrite /dev/acme/tag: %s\n",strerror(errno));
		close(fd);
	}
}

int
killchild(void)
{
	int fd;
	char *procfile;

	procfile = malloc(4096);
	sprintf(procfile, "/proc/%d/note",pid);


	if((fd = open(procfile,O_WRONLY)) > 0) {
		if(write(fd,"kill",4) != 4) {
			free(procfile);
			close(fd);
			return Error;
		}
		free(procfile);
		close(fd);
		return Success;
	}

	free(procfile);
	return Error;
}


int
execshellcmd(char *cmdline)
{
	int i = 0;
	char *cmd, **args;

	strtok(cmdline," ");

	cmd = malloc(strlen(cmdline)+5);
	sprintf(cmd,"/bin/%s",cmdline);

	args = malloc(512);
	args[0] = malloc(strlen(cmd));
	strcpy(args[0],cmd);

	cmdline = strtok(0," ");

	if(!cmdline) {

		if(!rfork(RFPROC | RFENVG | RFFDG | RFNAMEG)) {
			execv(cmd,args);
			printf("execv(); failed: %s\n",strerror(errno));
			return Error;
		}

		return Success;
	} else {


		for(; cmdline ;) {
			i++;
			args[i] = malloc(strlen(cmdline));
			args[i] = strdup(cmdline);

			cmdline = strtok(0, " ");

		}
		args[++i] = 0;

		if(!rfork(RFPROC | RFENVG | RFFDG | RFNAMEG)) {
			execv(cmd,args);
			printf("execv(); failed: %s\n",strerror(errno));
			return Error;
		}

		return Success;
	}

}

/****************************		playlist management	*************************************/


void
plsinit(void)
{
	plsroot = malloc(sizeof(Pls));

	plscurrent = plsroot;
	plslast = plsroot;
	plslast->next = 0;

	plsroot->artist = 0;
	plsroot->title = 0;
}


Pls *
plsclearregexec(Reprog *re, Pls *pls, char *tomatch, int not)
{
	Pls *pls2;

	if(pls == plsroot)
		pls = pls->next;

	switch( (regexec(re,tomatch,0,0) << 1) | not) {
		case 0:
		case 3:
			pls = pls->next;
			break;
		case 2:
		case 1:
			pls2 = pls;
			if(pls->next == 0) 
				pls->prev->next = 0;	
			else {
				pls->prev->next = pls->next;
				pls2->next->prev = pls->prev;
			}
			free(pls);
			pls = pls2->next;
	}

	return pls;
}

int
plsclear(int not, int vc, char *regex)
{
	Pls *pls, *pls2;
	Reprog *re = 0;
	int match = 0;
	char *tempchar;

	if(regex) {
		match = 1;

		if(!strcmp(regex,"-n")) {
			printf("NOT(s) without a regular expression. returning..\n");
			return Error;
		}

		if(!strcmp(regex,"-vc")) {
			printf("-vc without a regular expression. returning...\n");
			return Error;
		}
	}


	if(regex)
		re = regcomp(regex);

	pls = plsroot->next;

	while(pls) {

		switch( (match << 1) | vc) {
			case 0:	// vc 0 match 0
				pls2 = pls->next;
				free(pls);
				pls = pls2;
				break;
			case 2:	// vc 0 match 1
				pls = plsclearregexec(re,pls,pls->filename,not);
				break;
			case 3:		// vc 1 match 1
				if(pls->artist[0] == 0) {
					tempchar = strdup(pls->filename);
				} else {
					tempchar = malloc(strlen(pls->artist) + strlen(pls->title));
					sprintf(tempchar, "%s%s",pls->artist,pls->title);
				}

				pls = plsclearregexec(re,pls,tempchar,not);

				free(tempchar);
				break;
			default:
				printf("Well, I will better abort this infinte loop...\n");
				return Error;
		}
	}

	switch( (match << 1) | vc) {
		case 0:
		case 1:
			plsinit();
	}

	return Success;

}


int
plsgetnext(void)
{
	Pls *pls;
	int i,n;

	srand(time(0));

	if(shuffle) {
		pls = plsroot->next;

		for(i=0; pls ; i++)
			pls = pls->next;
	
		n = rand()%i;

		pls = plsroot->next;

		for(i = 0; i < n; i++)
			pls = pls->next;

		plscurrent = pls;

		return Success;

	}

	pls = plscurrent;

	if(pls->next) {

		plscurrent = pls->next;
		return Success;

	} else {

		plscurrent = plsroot->next;

		if(repeat) 
			return Success;

		return Error; // Not to be considered an error! Maybe make that '!Success' to make it look cleaner... .

	}
			
}

/* this function returns only for semantical reasons - always returns Success */
int
plswouldbenext(void)
{
	
	if(shuffle) {
		printf("I cannot guess - shuffle is on\n");
		return Success;
	}

	if(plscurrent == plslast) {
		if(repeat) {
			printf("repeating. Then, next file will be: %s\n",plsroot->next->filename);
			return Success;
		} else {
			switch(playmode) {
				case Pplay:
				case Ppause:
					printf("playlist ends after current song\n");
					return Success;
				case Pstop:
					printf("playlist ended. I will set it back to the beginning for you...\n");
					plscurrent = plsroot->next;
			}
		}
	}

	printf("next file will be: %s\n",plscurrent->filename);

	return Success;
}

/* We need to step two backwards, as we are fighting against plsgetnext() ... */
void
plsgetprev(void)
{
	if(plscurrent->prev != plsroot)
		plscurrent = plscurrent->prev->prev;
	else
		plscurrent = plsroot;
}


int
plsjump(int n)
{
	Pls *pls;
	int i;


	n--; // We are fighting against plsgetnext()...

	if(n == 0) { // n was 1. To start at track one, just set current to / and return.
		plscurrent = plsroot->next;
		return Success;
	}
	if(n < 0) {
		printf("negative integer.\n");
		return Error;
	}

	pls = plsroot->next;

	for(i = 0; pls ; i++)
		pls = pls->next;

	if(n > i) {
		printf("cannot jump to %d: There is no such song\n",n);
		return Error;
	}

	pls = plsroot->next;

	for(i = 0; i < n; i++)
		pls = pls->next;

	plscurrent = pls;

	return Success;
}


int
plsadd(char *filename)
{
	OggVorbis_File vf;
	FILE *fd;
	Pls *pls;
	int i;

	pls = malloc(sizeof(Pls));

	pls->filename = strdup(filename);
	pls->artist = malloc(1);
	pls->artist[0] = 0;
	pls->title = malloc(1);
	pls->title[0] = 0;


	if((fd = fopen(filename,"r")) == NULL) {
		printf("Could not open %s [%s]\n",filename,strerror(errno));
		return Error;
	}
	if(ov_open(fd, &vf, NULL, 0) < 0) {
		printf("ERROR: Failed to open input as vorbis\n");
		fclose(fd);
		return Error;
	}
	for(i = 0; i < vf.vc->comments;i++) {
		if(!strncmp(vf.vc->user_comments[i],"artist=",7))
			pls->artist = strdup(vf.vc->user_comments[i] + 7);

		if(!strncmp(vf.vc->user_comments[i],"title=",6))
			pls->title = strdup(vf.vc->user_comments[i] + 6);
	}


	pls->prev = plslast;
	plslast->next = pls;
	pls->next = 0;
	plslast = pls;

		
	

	if(plsroot == plscurrent) // virginity
		plscurrent = pls;
	else 
		if(shuffle) 
			plsgetnext();


	ov_clear(&vf);
	return Success;
				
}

void
plsprintfplaymode(void)
{
	printf("\t\t");
	switch(playmode) {
		case Ppause:
			printf("||\n");
			break;
		case Pplay:
			printf("|>\n");
			break;
		case Pstop:
			printf("[]\n");
	}
}

void
plsprintfregexec(Reprog *re, char *tomatch, Pls *pls, int not,int i)
{

	switch( (regexec(re,tomatch,0,0) << 1) | not) {
		case 0:
		case 3:
			break;
		case 2:
		case 1:
			printf("#  [%d]",i);

			if( pls->artist[0] && pls->title[0])
				printf(" %s - %s",pls->artist,pls->title);

			if(pls != plscurrent) 
				printf("\n");
			else 
				plsprintfplaymode();

			printf("%s\n",pls->filename);

	}

}


int
plsprintf(int not, int vc, char *regex)
{
	Pls *pls;
	Reprog *re = 0;	
	int i = 0,match = 0;
	char *tempchar;

	if(regex) {
		match = 1;

		if(!strcmp(regex,"-n")) {
			// 'pls -n -n' ?
			if(not) {
				printf("NOT(s) without a regular expression. returning..\n");
				return Error;
			} else {
				match = 0;
				regex = 0;
			}
		}

		if(!strcmp(regex,"-vc")) {
			match = 0;
			regex = 0;
		}
	}



	printf("shuffle: %s; repeat: %s; announcements: %s\n",
			shuffle ? "on" : "off",
			repeat ? "on" : "off",
			announcements ? "on" : "off");
	printf("Playmode: ");
	switch(playmode) {
		case Ppause:
			printf("paused\n");
			break;
		case Pplay:
			printf("playing\n");
			break;
		case Pstop:
			printf("stopped\n");
			break;
	}
	if(match)
		printf("Regexps: print entries with %s%smatching %s\n", 
				vc ? "vorbis-comments" : "filenames", 
				not ? " not " : " ", 
				regex);


	if(plslast == plsroot) {
		printf("[empty playlist]\n");
		return Success;
	}


	pls = plsroot->next;

	if(regex)
		re = regcomp(regex);

	while(pls) {
		i++;

		switch( (match << 1) | vc) {
			case 0:		// vc 0 match 0
			case 1:		// vc 1 match 0
			
				printf("#  [%d]",i);

				if( pls->artist[0] && pls->title[0])
					printf(" %s - %s",pls->artist,pls->title);

				if(pls != plscurrent) 
					printf("\n");
				else 
					plsprintfplaymode();

				printf("%s\n",pls->filename);
				break;
			case 2:		// vc 0 match 1
				plsprintfregexec(re,pls->filename,pls,not,i);
				break;
			case 3:		// vc 1 match 1
				tempchar = malloc(strlen(pls->artist) + strlen(pls->title));
				sprintf(tempchar,"%s%s",pls->artist,pls->title);

				plsprintfregexec(re,tempchar,pls,not,i);

				free(tempchar);
				break;
			default:
				printf("plsprintf(): error in application logic\n");
				return Error;
		}
	
		pls = pls->next;
	}

	return Success;

}



int
plsdump(char *filename)
{
	Pls *pls;
	int fd;
	char *tempchar;

	if((fd = open(filename,O_WRONLY | O_TRUNC | O_CREAT)) > 0) {

		pls = plsroot->next;

		tempchar = malloc(4096);

		while(pls) {

			if( pls->artist[0]  && pls->title[0] ) {
				memset(tempchar,0,4096);
				sprintf(tempchar,"# %s - %s\n",pls->artist,pls->title);
				if(write(fd,tempchar,strlen(tempchar)) != strlen(tempchar)) {
					printf("Could not write to %s\n",filename);
					close(fd);
					free(tempchar);
					return Error;
				}
			}

			if(write(fd,pls->filename,strlen(pls->filename)) != strlen(pls->filename)) {
				printf("Could not write to %s\n",filename);
				close(fd);
				free(tempchar);
				return Error;
			}
			if(write(fd,"\n",1) != 1) {
				printf("Could not write to %s\n",filename);
				close(fd);
				free(tempchar);
				return Error;
			}

			pls = pls->next;
		}

		printf("dumped: %s\n",filename);	// for plumbing...
		free(tempchar);
		close(fd);
		return Success;
	}

	printf("Could not open %s: %s\n",filename,strerror(errno));
	return Error;
}

/***********************************	audio	***********************************************/


int
chvol(int vol)
{
	int fd;
	char *cmd;

	cmd = malloc(4096);
	sprintf(cmd,"audio out %d\n",vol);

	if((fd = open("/dev/volume",O_WRONLY)) > 0) {
		write(fd,cmd,strlen(cmd));
		free(cmd);
		close(fd);
		return Success;
	}

	printf("Could not open /dev/volume\n");

	free(cmd);
	return Error; 
}


int
isogg(char *file)
{
	OggVorbis_File vf;
	FILE *fd;

	if((fd = fopen(file,"r")) == NULL) {
		printf("Could not open %s: %s\n",file,strerror(errno));
		return Error;
	}

	if(ov_open(fd, &vf, NULL, 0) < 0) {
		printf("ERROR: Failed to open input as vorbis\n");
		return Error;
	}


	ov_clear(&vf);
	return Success;
}

void
chaudiorate(long newrate)
{
	int fd;
	char *cmd;

	cmd = malloc(4096);
	sprintf(cmd,"speed %d\n",newrate);

	if((fd = open("/dev/volume",O_WRONLY)) > 0) {
		write(fd,cmd,strlen(cmd));
		free(cmd);
		close(fd);
		audiorate = newrate;
	} else {
		printf("Could not open /dev/volume: %s\n", strerror(errno));
		exit(1);
	}
}


void
getsamplerate(void)
{
	int fd;
	char buffer[512];

	if((fd = open("/dev/volume",O_RDONLY)) > 0) {
		while(read(fd,buffer,512) != 0) {
			if(!strncmp(buffer,"speed in out ",13))
				audiorate = atol(buffer + 13);
		}
		close(fd);
	} else {
		printf("getsamplerate(): Could not open /dev/volume for reading out the sample-rate\n");
		audiorate = 44100;
	}
}

int
playvorbis(char *file)
{
	OggVorbis_File vf;
	FILE *fd;
	ogg_int64_t length = 0;
	int afd, ret, bs = 0;
	unsigned int written = 0;
	char buffer[8192];


	USED(written);


	if((fd = fopen(file,"r")) == NULL) {
		printf("Could not open %s for reading.\n",file);
		return Error;
	}

	if((afd = open("/dev/audio",O_WRONLY)) <= 0) {
		printf("playvorbis(): Could not open /dev/audio for writting (fd = %d)\n",afd);
		return Error;
	}


	if(ov_open(fd, &vf, NULL, 0) < 0) {
		printf("ERROR: Failed to open input as vorbis\n");
		fclose(fd);
		return Error;
	}

	// TODO: Why this?
 	if(ov_seekable(&vf)) 
		length = ov_pcm_total(&vf, 0);
	

	/* Right number of samples per second for audio(3) ? */
	if(vf.vi->rate != audiorate)
		chaudiorate(vf.vi->rate);


	/* Right number of channels for audio(3) ? */
	 if(vf.vi->channels != 2) {
		printf("playvorbis: Channels != 2 - unsupported, skipped\n");
		ov_clear(&vf);
		close(afd);
		fclose(fd);
		return Error;
	}


			

	/* Main decode/ audio-out  loop */
	while((ret = ov_read(&vf, buffer, 8192, 0, 2, 1, &bs)) != 0) {
		if(bs != 0) {
			printf("Only one logical bitstream currently supported\n");
			break;
		}

		if(ret < 0 ) {
            		   printf("Warning: hole in data\n");
			continue;
		}

		
		if(skipper != 0) {
			double s = ov_time_tell(&vf);

			s += skipper;
			if(s < 0.0)
				s = 0;

			ov_time_seek(&vf,s);


			skipper = 0;
			continue;
		}

		if(whattimeisit) {
			printf("current position in bitstream: %f\n",ov_time_tell(&vf));
			whattimeisit = !whattimeisit;
		}

		playmodeswitching:

		switch(playmode) {
			case Pstop:
				ov_clear(&vf);
				close(afd);
				fclose(fd);
				return Success;
			case Pplay:
				if(write(afd, buffer, ret) != ret) {
					printf("Error writting audio out\n");
					ov_clear(&vf);
					close(afd);
					fclose(fd);
					return Error;
				}
				break;
			case Ppause:
				sleep(3);
				goto playmodeswitching;
		}

		written += ret;
       
	}


	fclose(fd);
	close(afd);
	ov_clear(&vf);
	return Success;
}

int
play(void)
{

	if(plsroot == plslast) {
		printf("empty playlist\n");
		return Error;
	}

	/* As we don't know if this is the first call, we simply get the rate */
	getsamplerate();

	while(1) {
		if(announcements)
			announcer();

		if(playvorbis(plscurrent->filename) == Error)
			return Error;

		if(playmode == Pstop)	// We may be here because playmode was switched.
			return Success;

		if(plsgetnext() == Error) {
			printf("no next song, assuming end of playlist.\n");
			break;
		}

	}

	return Success;

}



/************************************	misc		**********************************************/


void
usage(void)
{
	printf("play\t\t\t\t-- start playback\n");
	printf("pause\t\t\t-- stop playback-thread\n");
	printf("stop\t\t\t\t-- kill playback-thread\n");
	printf("shuffle\t\t\t-- switch on shuffle (shuffle also is repeat - nobody knows when to stop)\n");
	printf("repeat\t\t\t-- switch on repeat\n");
	printf("announce\t\t\t-- switch off anouncements\n");
	printf("np\t\t\t\t-- printf what song is currently played\n");
	printf("skip N\t\t\t-- fast forward/backward N seconds\n");
	printf("next\t\t\t\t-- skips current track (if playback is paused or stopped, next restarts playback)\n");
	printf("prev\t\t\t\t-- steps back to previous track (if playback is paused or stopped, prev restarts playback)\n");
	printf("wbnext\t\t\t--WouldBenext - show what track comes next\n");
	printf("jump N\t\t\t--jump to track-nr. N\n");
	printf("clear [/regexp/]\t-- clear playlist (optionally clear only files matching regexp. note: If you want to match on Vorbis-comments, use clear -vc /regexp/ ; use -n to NOT your selection)\n");
	printf("pls [...]\t\t\t--Prints current playlist. Supports -n and -vc, too\n");
	printf("exit | quit\t\t\t-- exit 9vorbis normally. Child is killed.\n");
	printf("abort\t\t\t-- exit 9vorbis without taking care. See the Abort-section below.\n");
	printf("vol N\t\t\t-- Sets the output volume to N (n > -1 < 101)\n");
	printf("bl0b.ogg\t\t\t-- put bl0b.ogg into the playlist.\n");
	printf("dump [filename]\t-- dump current playlist to filename (default: $home/dump.pls)\n");
	printf("open [filename]\t-- opens commandset from filename (default: $home/dump.pls)\n");
	printf("help\t\t\t\t-- a help screen\n");
}


/* Is it me or is this bitch especially slow? I demand libbio in APE :P */
int
cmdsetopen(char *filename)
{
	FILE *fd;
	char buffer[4096];

	if((fd = fopen(filename,"r")) == NULL) {
		printf("Could not open '%s' for reading\n",filename);
		return Error;
	}



	while(fgets(buffer,sizeof(buffer),fd) != NULL) {
		if(strstr(buffer,"\n") != NULL) 
			buffer[strlen(buffer)-1] = 0;

		if(execcmd(buffer) == Error) 
			errorcount++;

	}

	fclose(fd);
	return Success;
}

int
execcmd(char *cmd)
{
	int tempi;


	if(!strcmp(cmd,""))
		return Success; // ignore blanks

	else if(!strncmp(cmd,"9vorbis> ",7)) {		// ec_9vorbis

		cmd += 7;

		return execcmd(cmd);

	} else if(!strncmp(cmd,"#",1)) {			// ec_#

		return Success;

	} else if(!strncmp(cmd,"play",4)) {		// ec_play

		switch(playmode) {
			case Pstop:
				break;
			case Ppause:
				playmode = Pplay;
				return Success;
			case Pplay:
				playmode = Pstop;
				wait(&tempi);
				return execcmd("play");
			default:
				printf("playmode %d in execcmd(\"play\");\n",playmode);
				return Error;
		}

		playmode = Pplay;

		switch((pid = rfork(RFMEM | RFPROC | RFNAMEG))) {
			case 0:
				errorcount += play();
				playmode = Pstop; // never becomes Ppause when play() returns...
				exit(0);
			case -1:
				playmode = Pstop;
				printf("I could not fork\n");
				return Error;
		}
		return Success;

	} else if(!strncmp(cmd,"pause",5)) {	// ec_pause

		switch(playmode) {
			case Pstop:
				return Success;
			case Ppause:
				playmode = Pplay;
				return Success;
			case Pplay:
				playmode = Ppause;
				return Success;
			default:
				printf("playmode %d in execcmd(\"play\");\n",playmode);
				return Error;
		}

	} else if(!strncmp(cmd,"stop",4)) {	// ec_stop

		playmode = Pstop;
		return Success;

	} else if(!strncmp(cmd,"kill",4)) {	// ec_kill

		playmode = Pstop;
		return killchild();

	} else if(!strncmp(cmd,"shuffle",8)) {	// ec_shuffle

		printf("switching shuffle %s\n", shuffle ? "off" : "on");
		shuffle = !shuffle;
		return Success;

	} else if(!strncmp(cmd,"repeat",6)) {	// ec_repeat

		printf("switching repeat %s\n", repeat ? "off" : "on");
		repeat = !repeat;
		return Success;

	} else if(!strncmp(cmd,"announcements",7)) {	// ec_announcements

		printf("switching announcements %s\n", announcements ? "off" : "on");
		announcements = !announcements;
		return Success;

	} else if(!strncmp(cmd,"np",2)) {	// ec_np

		announcer();
		return Success;

	} else if(!strncmp(cmd,"skip",4)) {	// ec_skip

		cmd += 4;

		if(*cmd == '\0') {
			whattimeisit = 1;
			return Success;
		}
		cmd++;

		if(*cmd == '\0') {
			printf("syntax error: skip: expected anything, got end of string\n");
			return Error;
		}

		tempi = atoi(cmd);

		skipper += tempi;

		return Success;

	} else if(!strncmp(cmd,"next",4)) {	// ec_next

		switch(playmode) {
			case Ppause:
			case Pplay:
				playmode = Pstop;
				wait(&tempi);
		}

		if(plsgetnext() == Error) {
			printf("no next song, assuming end of playlist.\n");
			return Error;
		}

		return execcmd("play");

	} else if(!strncmp(cmd,"wbnext",6)) {	// ec_wbnext

		return plswouldbenext();

	} else if(!strncmp(cmd,"prev",4)) {		// ec_prev

		
		switch(playmode) {
			case Ppause:
			case Pplay:
				playmode = Pstop;
				wait(&tempi);
		}
		plsgetprev();
		return execcmd("play");

	} else if(!strncmp(cmd,"jump ",5)) {	// ec_jump

		strtok(cmd," ");
		cmd = strtok(0," ");


		switch(playmode) {
			case Ppause:
			case Pplay:
				playmode = Pstop;
				wait(&tempi);
		}

		tempi = atoi(cmd);

		if(tempi == 0) {
			printf("ASCII to integer conversion failed or you try to jump to 0\n");
			return Error;
		}


		if(plsjump(tempi) == Error)
			return Error;

		return execcmd("play");

	} else if(!strncmp(cmd,"!",1)) {	// ec_!

		return execshellcmd(cmd+1);

	} else if(!strncmp(cmd,"pls",3)) {	// ec_pls

		strtok(cmd," ");
		cmd = strtok(0," ");

		if(!cmd)
			return plsprintf(0,0,0);
		else {
			int not = 0, vc = 0;
			char *regex = 0;

			for(; cmd ;) {
				if(!strcmp(cmd,"-n"))
					not = !not;
				if(!strcmp(cmd, "-vc"))
					vc = 1;
				regex = cmd;
				cmd = strtok(0, " ");
			}

			return plsprintf(not,vc,regex);
		}

	} else if(!strncmp(cmd, "clear",5)) {	// ec_clear

		strtok(cmd," ");
		cmd = strtok(0, " ");

		if(!cmd)
			return plsclear(0,0,0);
		else {
			int not = 0, vc = 0;
			char *regex = 0;

			for(; cmd ;) {
				if(!strcmp(cmd,"-n"))
					not = !not;
				if(!strcmp(cmd, "-vc"))
					vc = 1;
				regex = cmd;
				cmd = strtok(0, " ");
			}

			return plsclear(not,vc,regex);
		}

	} else if(!strncmp(cmd,"exit",4)) {	// 	ec_exit


		killchild();

		if(errorcount > 0)
			printf("Errorcount: %d\n",errorcount);

		exit(0);

	} else if(!strncmp(cmd,"quit",4)) {	// ec_quit

		execcmd("exit");

	} else if(!strncmp(cmd,"abort",5)) {	// ec_abort

		exit(0);

	} else if(!strncmp(cmd,"vol",3)) {	// ec_vol

		strtok(cmd," ");
		cmd = strtok(0, " ");

		if(cmd)
			return chvol(atoi(cmd));
		else {
			printf("!cat /dev/volume\n");
			return Success;
		}
		

	} else if(!strncmp(cmd,"dump",4)) {	// ec_dump

		strtok(cmd," ");
		cmd = strtok(0, " ");

		if(!cmd) {
			char *tempchar;
			tempchar = malloc(1024);
			sprintf(tempchar,"%s/dump.pls",getenv("home"));
			return plsdump(tempchar);
		} else
			return plsdump(cmd);

	} else if(!strncmp(cmd,"open",4)) {	// ec_open

		strtok(cmd," ");
		cmd = strtok(0, " ");

		if(!cmd) {
			char *tempchar;
			tempchar = malloc(1024);
			sprintf(tempchar,"%s/dump.pls",getenv("home"));
			return cmdsetopen(tempchar);
		} else
			return cmdsetopen(cmd);

	} else if(!strncmp(cmd,"help",4)) {	// ec_help

		usage();
		return Error; // n00b

	} else {	// ec_isogg ec_plsadd

		if(isogg(cmd) == Error) {	
			printf("Errors adding '%s'\n",cmd);
			return Error;
		}
		// isogg() protects us from dirs, non-regular files and broken ogg-vorbis-files
		plsadd(cmd);
	}

	return Success;
}



void
main(int argc, char **argv)
{
	int i;
	char *buffer;

	plsinit();

	if(argc > 1) {
		for(i = 1;i < argc;i++) {
			printf("9vorbis> %s\n",argv[i]);
			if(execcmd(argv[i]) == Error) {
				errorcount++;
			}
		}
	}


	buffer = malloc(4096);

	prepacme();

/*
There must be a reason why 
% 9vorbis <bl0b.pls
makes this loop go wild - but I have not found
out yet. You're very welcome.
*/

mainloop:
	printf("9vorbis> ");
	fflush(stdout);

	while(read(0,buffer,4096)) {
		buffer[strlen(buffer)-1] = '\0';

	
		if(execcmd(buffer) == Error)
			errorcount++;

		memset(buffer,0,4096);

		printf("9vorbis> ");
		fflush(0);
	}
	printf("\n");
	fflush(stdout);
	goto mainloop;
}
